//
//  Copyright (C) 2012 Vitor Rolla
//  vitorgr [at] dei.uc.pt
//  University of Coimbra - Portugal
//

#include "MOFO_Timer.h"
#include <stdexcept>
#include <sstream>

namespace MOFO{

MOFO_Timer::MOFO_Timer(cSimpleModule* parent, std::string name, simtime_t interval) : parent(parent), interval(interval), active(false)
{
    this->name = strdup(name.c_str());
    message = new MOFO_Timeout(this->name, 1 /* msg kind, to give the msg a green color */);
}

MOFO_Timer::~MOFO_Timer()
{
    cancel();
    delete message;
    free(this->name);
}

const char* MOFO_Timer::getFullName() const
{
    return name;
}

std::string MOFO_Timer::info() const
{
    if (!active)
    {
        return "inactive";
    }
    std::ostringstream ss;
    ss << (expiresAt - (simTime())) << "s left";
    return ss.str();
}

std::string MOFO_Timer::detailedInfo() const
{
    return info();
}

bool MOFO_Timer::isRunning()
{
    if (active && (expiresAt > simTime()))
    {
        return true;
    }
    return false;
}

bool MOFO_Timer::isExpired()
{
    if (active && (expiresAt <= simTime()))
    {
        parent->cancelEvent(message);
        active = false;
        return true;
    }
    return false;
}

void MOFO_Timer::start(simtime_t interval)
{
    if (interval != 0) this->interval = interval;
    if (this->interval == 0) throw std::runtime_error("Tried starting MOFO_Timer without or with zero-length interval");
    if (active) parent->cancelEvent(message);
    expiresAt = simTime() + this->interval;
    active = true;
    parent->scheduleAt(expiresAt, message);
}

void MOFO_Timer::cancel()
{
    if (active) parent->cancelEvent(message);
    active = false;
}

bool MOFO_Timer::owns(const cMessage* message) const
{
    return (message == this->message);
}

simtime_t MOFO_Timer::getInterval() const
{
    return interval;
}

std::ostream& operator<< (std::ostream& os, const MOFO_Timer& o)
{
    os << o.info();
    return os;
}

}
