//
//  Copyright (C) 2012 Vitor Rolla
//  vitorgr [at] dei.uc.pt
//  University of Coimbra - Portugal
//

#include "MOFO.h"
#include "IPProtocolId_m.h"
#include "ModuleAccess.h"
#include "IP.h"
#include "IPControlInfo.h"



Define_Module( MOFO_Net );

namespace {
	const int MOFO_CTRL_HEADER_LENGTH = 13; /**< length (in bytes) of a MOFO RM header */
	const double VALOR_DA_MSG = 1; /**< length (in bytes) of a MOFO RM header */
	const int MOFO_MBLOCK_LENGTH = 10; /**< length (in bytes) of one MOFO RBlock */
	const double MAXJITTER = 0.001; /**< all messages sent to a lower layer are delayed by 0..MAXJITTER seconds (draft-ietf-manet-jitter-01) */
	const IPAddress LL_MANET_ROUTERS = "224.0.0.90"; /**< Link-local multi-cast address of all MANET routers (TBD) */
}

void MOFO_Net::initialize(int aStage)
{
    cSimpleModule::initialize(aStage);

    if (0 == aStage)
    {

    	mUppergateIn  = findGate("uppergate$i");
		mUppergateOut = findGate("uppergate$o");
		mLowergateIn[0]  = findGate("ifIn", 0);
		mLowergateOut[0] = findGate("ifOut", 0);
		mLowergateIn[1]  = findGate("ifIn", 1);
		mLowergateOut[1] = findGate("ifOut", 1);


    	//NED File Parameters
		RESPONSIBLE_ADDRESSES_PREFIX=par("RESPONSIBLE_ADDRESSES_PREFIX");
		MOFO_INTERFACES=par("MOFO_INTERFACES");
		AUTOASSIGN_ADDRESS_BASE=IPAddress(par("AUTOASSIGN_ADDRESS_BASE").stringValue());
		BUFFER_SIZE_PACKETS=par("BUFFER_SIZE_PACKETS");
		BUFFER_SIZE_BYTES=par("BUFFER_SIZE_BYTES");
		CTRL_MSG_PERIOD=par("CTRL_MSG_PERIOD");
		NUM_OF_COPIES=par("NUM_OF_COPIES");
		LAMBDA=par("LAMBDA");



    	ctrlTimeout = new MOFO_Timer(this, "ctrlMsgTimeout", CTRL_MSG_PERIOD);
    	WATCH_PTR(ctrlTimeout);

    	//Initialize MOFO stats
    	totalPacketsReplicated = 0;
    	WATCH(totalPacketsReplicated);
    	totalPacketsRcvd = 0;
    	WATCH(totalPacketsRcvd);
        totalPacketsSent = 0;
        WATCH(totalPacketsSent);
        totalBytesSent = 0;
        WATCH(totalBytesSent);
        totalPacketsDeleted = 0;
        WATCH(totalPacketsDeleted);
        totalSelfishPacketsDelivery = 0;
		WATCH(totalSelfishPacketsDelivery);

		selfishIsoDelivery = 0;
		WATCH(selfishIsoDelivery);
		unselfishIsoDelivery = 0;
		WATCH(unselfishIsoDelivery);

        totalRcvdReward= 0;
		WATCH(totalRcvdReward);


        statsCTRLSent = 0; /**< number of generated MOFO CTRL MSG */
        statsCTRLRcvd = 0; /**< number of consumed MOFO CTRL MSG */
        statsRQSTSent = 0; /**< number of generated MOFO RQST MSG */
        statsRQSTRcvd = 0; /**< number of consumed MOFO RQST MSG */
        statsACKSent  = 0;
        statsACKRcvd  = 0;


        dataLatencySelfish = 0;
        dataLatencyUnselfish = 0;
        dataSamplesUnselfish = 0;
        dataSamplesSelfish = 0;

        //rblockSize = 13;

        bufferValueVec.setName("Total Buffer Value");
        numOfNonSelfishPacketsOnBuffer.setName("Num Of Non-Selfish Packets on Buffer");
        nOfHopsVector.setName("Number of Hops");

        dataDelayVec.setName("Data delay");
        //dataLoadVec.setName("Data load");
        //controlLoadVec.setName("Control load");

        // -v- MOFOnet needs AUTOASSIGN_ADDRESS_BASE to be set to 0.0.0.0
		if (AUTOASSIGN_ADDRESS_BASE.getInt() != 0) error("MOFOnet needs AUTOASSIGN_ADDRESS_BASE to be set to 0.0.0.0");
		// -^-

		myAddr = AUTOASSIGN_ADDRESS_BASE.getInt() + uint32(getParentModule()->getId());
		queuedElements = new cQueue();
		numQueuedMsgs = 0;

       	}

       	if (2 == aStage) {
       		ctrlTimeout->start(CTRL_MSG_PERIOD);
       	}
}

void MOFO_Net::finish()
{
	recordScalar("totalPacketsReplicated", totalPacketsReplicated);
	recordScalar("totalPacketsRcvd", totalPacketsRcvd);
    recordScalar("totalPacketsSent", totalPacketsSent);
    recordScalar("totalPacketsDeleted", totalPacketsDeleted);
    recordScalar("totalBytesSent", totalBytesSent);
    recordScalar("totalRcvdReward", totalRcvdReward);
    recordScalar("totalSelfishPacketsDelivery", totalSelfishPacketsDelivery);
    //recordScalar("statsNodeDensity", statsNodeDensity);
    recordScalar("selfishIsoDelivery", selfishIsoDelivery);
    recordScalar("unselfishIsoDelivery", unselfishIsoDelivery);

    recordScalar("MOFO_CTRLSent", statsCTRLSent);
    recordScalar("MOFO_CTRLRcvd", statsCTRLRcvd);
    recordScalar("MOFO_RQSTSent", statsRQSTSent);
    recordScalar("MOFO_RQSTRcvd", statsRQSTRcvd);
    recordScalar("MOFO_ACKSent", statsACKSent);
    recordScalar("MOFO_ACKRcvd", statsACKRcvd);


    //recordScalar("MOFO_MOFORcvd", statsMOFORcvd);

    //recordScalar("MOFO_NodeDensity", statsNodeDensity);



    //if (meanBufferValue > 0 && bufferSamples > 0)
        //recordScalar("buffer value", meanBufferValue/bufferSamples);

    if (dataLatencyUnselfish > 0 && dataSamplesUnselfish > 0)
            recordScalar("data latency unselfish", dataLatencyUnselfish/dataSamplesUnselfish);

    if (dataLatencySelfish > 0 && dataSamplesSelfish > 0)
                recordScalar("data latency selfish", dataLatencySelfish/dataSamplesSelfish);


	while (! queuedElements->isEmpty()) {
		QueueElement *el = static_cast<QueueElement *>(queuedElements->remove(queuedElements->get(0)));
		delete el->getObject();
		delete el;
		//totalPacketsDeleted++;
	}
	delete queuedElements;


}

void MOFO_Net::handleMessage(cMessage* apMsg) {
	if (mUppergateIn == apMsg->getArrivalGateId()) {
		handleUpperMsg(check_and_cast<cPacket*>(apMsg));
	}
	else if (apMsg->isSelfMessage()) {
		handleSelfMsg(apMsg);
	}
	else {
		cPacket* apPkt = check_and_cast<cPacket*>(apMsg);
		handleLowerMsg(apPkt);
	}
}



/* ------- Process Incoming UpperDataMsg ------- */

void MOFO_Net::handleUpperMsg(cPacket* apMsg) {
	MOFO_Packet * pkt = encapsMsg(apMsg);

	if(pkt->getDestAddress() == myAddr)
	{
		sendUp(pkt->decapsulate());
		delete pkt;
		// -^-

		return;
	}

	insertMsg(pkt);
}




MOFO_Packet * MOFO_Net::encapsMsg(cPacket * apMsg) {

	AppPkt * control_info = check_and_cast<AppPkt *>(apMsg->removeControlInfo());
	MOFO_Packet * mofoPkt = new MOFO_Packet(apMsg->getName());

	cModule* destinationModule = getParentModule()->getParentModule()->getModuleByRelativePath(control_info->getDestName());
	if(!destinationModule) error("destination module not found");
	//int msgId = (int) control_info->par("msgId");

	/** get the IP-Address of the destination node **/
	unsigned int destAddr = destinationModule->getId();

	mofoPkt->setDestAddress(destAddr);
	mofoPkt->setSrcAddress(myAddr);
	mofoPkt->setCreationTime(simTime());
	mofoPkt->setMsgId(control_info->getMsgId());
	mofoPkt->setMsgId(control_info->getMsgId());
	//NUM_OF_COPIES = (rand() % 25)+24;
	//NUM_OF_COPIES = 50;
	mofoPkt->setNOfHops(0);
	mofoPkt->setNOfCopies(NUM_OF_COPIES);
	mofoPkt->setInitNOfCopies(NUM_OF_COPIES);
	mofoPkt->encapsulate(apMsg);
	delete control_info;
	totalPacketsSent++;
	return mofoPkt;
}



void MOFO_Net::insertMsg(MOFO_Packet *packet){
	if (numQueuedMsgs >= BUFFER_SIZE_PACKETS){
		cQueue::Iterator queueIterator(*queuedElements);
		QueueElement * qElement1;
		simtime_t now = simTime();
		//double pktValue = (VALOR_DA_MSG/packet->getInitNOfCopies()*exp(LAMBDA*(SIMTIME_DBL(now) - SIMTIME_DBL(packet->getCreationTime()))));

		double minLocalMsgValue = VALOR_DA_MSG;

		double bufferValue = 0;

		int numOfNonSelfishPackets = 0;

		while(!queueIterator.end()){
			qElement1 = dynamic_cast<QueueElement*>(queueIterator());
			if (qElement1) {
			double localMsgValue = (VALOR_DA_MSG/log2(qElement1->getInitNumberOfCopies())*exp(LAMBDA*(SIMTIME_DBL(now) - SIMTIME_DBL(qElement1->getCreationTime()))));
			bufferValue = bufferValue + localMsgValue;
			if (localMsgValue < minLocalMsgValue) minLocalMsgValue = localMsgValue;
			}
			queueIterator++;
			if (qElement1->getInitNumberOfCopies() < 16)
					numOfNonSelfishPackets++;
		}
		bufferValueVec.record(bufferValue);
		numOfNonSelfishPacketsOnBuffer.record(numOfNonSelfishPackets);

		queueIterator.init(*queuedElements, false);

		while(!queueIterator.end()){
			qElement1 = dynamic_cast<QueueElement*>(queueIterator());
			if ((qElement1) && (minLocalMsgValue == (VALOR_DA_MSG/log2(qElement1->getInitNumberOfCopies())*exp(LAMBDA*(SIMTIME_DBL(now) - SIMTIME_DBL(qElement1->getCreationTime())))))) {
				ev << "Erasing data packet " << qElement1->getObject() << endl;
				//dataQueueByteSize -= qd.datagram->getByteLength();
				queuedElements->remove(qElement1);
				numQueuedMsgs--;
				totalPacketsDeleted++;
				break;
			}
			queueIterator++;
		}
	} else {
		cQueue::Iterator queueIterator(*queuedElements);
		QueueElement * qElement1;
		simtime_t now = simTime();
		double bufferValue = 0;
		int numOfNonSelfishPackets = 0;
		while(!queueIterator.end()){
			qElement1 = dynamic_cast<QueueElement*>(queueIterator());
			if (qElement1) {
				double localMsgValue = (VALOR_DA_MSG/log2(qElement1->getInitNumberOfCopies())*exp(LAMBDA*(SIMTIME_DBL(now) - SIMTIME_DBL(qElement1->getCreationTime()))));
			bufferValue = bufferValue + localMsgValue;
			//if (localMsgValue < minLocalMsgValue) minLocalMsgValue = localMsgValue;
			}
			queueIterator++;
			if (qElement1->getInitNumberOfCopies() < 16)
					numOfNonSelfishPackets++;
		}
		bufferValueVec.record(bufferValue);
		numOfNonSelfishPacketsOnBuffer.record(numOfNonSelfishPackets);
	}

	ev << "Queueing data packet " << packet << endl;
	if(packet->getSrcAddress() == myAddr)
		{
	QueueElement * qElement = new QueueElement(packet->getMsgId(), packet->getCreationTime(), packet->getDestAddress(), myAddr, packet->getNOfHops(), packet->getInitNOfCopies(), packet->getNOfCopies(), packet);
	queuedElements->insert(qElement);
		}
	else{
		//packet->setNOfCopies(1); //Switch: Two-hop vs. SnW
		QueueElement * qElement = new QueueElement(packet->getMsgId(), packet->getCreationTime(), packet->getDestAddress(), myAddr, packet->getNOfHops(), packet->getInitNOfCopies(), packet->getNOfCopies(), packet);
		queuedElements->insert(qElement);
	}
	messageVector.push_back(packet->getMsgId());
	//dataQueueByteSize += data->getByteLength();
	numQueuedMsgs++;

}



void MOFO_Net::handleLowerMsg(cPacket* apMsg) {
	// detach and free associated ControlInfo object
	delete apMsg->removeControlInfo();

	/**
	 * check the type of received message
	 1) CTRL
	 2) RQST
	 3) Data Message
	 4) ACK
	 **/
	if(dynamic_cast<MOFO_CTRL*>(apMsg)) handleLowerCTRL(dynamic_cast<MOFO_CTRL*>(apMsg));
	else if(dynamic_cast<MOFO_RQST*>(apMsg)) handleLowerRQST(dynamic_cast<MOFO_RQST*>(apMsg));
	else if(dynamic_cast<MOFO_Packet*>(apMsg)) handleLowerPacket(dynamic_cast<MOFO_Packet*>(apMsg));
	else if(dynamic_cast<MOFO_ACK*>(apMsg)) handleLowerACK(dynamic_cast<MOFO_ACK*>(apMsg));
	else error("cannot cast message to MOFO Packet");
}

void MOFO_Net::handleLowerACK(MOFO_ACK *rewardMsg) {

	ev << "received message is a ACK message" << endl;

	statsACKRcvd++;

	totalRcvdReward = totalRcvdReward + rewardMsg->getReward();

	delete rewardMsg;







}

void MOFO_Net::handleLowerPacket(MOFO_Packet *datagram) {

	IPAddress destAddr = datagram->getDestAddress();

	ev << "inbound data packet signalled" << endl;

	// if this is for us, record statistics
	if (destAddr.getInt() == myAddr) {

		std::vector<unsigned int>::iterator iter;

		for (iter = messageVector.begin(); iter < messageVector.end(); iter++)
		{
			if (*iter == datagram->getMsgId())
			{
				EV << "Duplicated packet: ";
				delete datagram;
				return;
			}
		}

		messageVector.push_back(datagram->getMsgId());
	    EV << "Received packet: ";

	    dataDelayVec.record(simTime() - datagram->getCreationTime());
	    if (datagram->getInitNOfCopies() < 16){
	    	dataLatencyUnselfish += SIMTIME_DBL(simTime() - datagram->getCreationTime());
	    	dataSamplesUnselfish++;
	    }
		if (datagram->getInitNOfCopies() > 8){
			dataLatencySelfish += SIMTIME_DBL(simTime() - datagram->getCreationTime());
			dataSamplesSelfish++;
		}
		datagram->setNOfHops(datagram->getNOfHops()+1);
		nOfHopsVector.record(datagram->getNOfHops());

		ev << "start buliding msg ACK" << endl;
		MOFO_ACK *my_ack = new MOFO_ACK("ACK");
		my_ack->setOrigNode(myAddr);
		double msgReward = (VALOR_DA_MSG/log2(datagram->getInitNOfCopies())*exp(LAMBDA*(SIMTIME_DBL(simTime()) - SIMTIME_DBL(datagram->getCreationTime()))));
		my_ack->setReward(msgReward);
		sendDown(my_ack, datagram->getMiddleAddress());
		statsACKSent++;

		if (datagram->getInitNOfCopies() > 8){
			totalSelfishPacketsDelivery++;
			if (datagram->getNOfHops() == 5.0)
				selfishIsoDelivery++;
		}

		if (datagram->getInitNOfCopies() < 16){
			if (datagram->getNOfHops() == 4.0)
				unselfishIsoDelivery++;
		}

		sendUp(datagram->decapsulate());


		delete datagram;
		// -^-

		return;
	}

	std::vector<unsigned int>::iterator iter;
	for (iter = messageVector.begin(); iter < messageVector.end(); iter++)
	{
		if (*iter == datagram->getMsgId())
		{
			EV << "Duplicated packet: ";
			delete datagram;
			return;
		}
	}
	datagram->setNOfHops(datagram->getNOfHops()+1);
	//messageVector.push_back(datagram->getMsgId());
	insertMsg(datagram);
	return;



	//delete datagram;
	// -^-
}

void MOFO_Net::sendUp(cPacket* apMsg) {
	ev << "sending up " << apMsg << "\n";

	totalPacketsRcvd++;
	send(apMsg, mUppergateOut);
}

void MOFO_Net::handleLowerRQST(MOFO_RQST *routingMsg)
{
	ev << "received message is a RQST message" << endl;

	statsRQSTRcvd++;

	ev << "starting update routes from routing blocks in the received message" << endl;
	std::vector<MOFO_MessageBlock> additional_nodes = routingMsg->getRqstMsgTable();

	for (unsigned int i = 0; i < additional_nodes.size(); i++){
		cQueue::Iterator queueIterator(*queuedElements);
		QueueElement * qElement;
		while(!queueIterator.end()){
			qElement = dynamic_cast<QueueElement*>(queueIterator());
			if ((qElement)&&((additional_nodes[i].getMsgId()) == (qElement->getMsgId())&&(qElement->getNumberOfCopies()>=2))) {
				totalPacketsReplicated++;
				qElement->setNumberOfCopies();
				if (qElement->getDestAddr() == routingMsg->getOrigNode()){
					qElement->getObject()->setMiddleAddress(myAddr);
					sendDown(qElement->getObject()->dup(), routingMsg->getOrigNode());
					queuedElements->remove(qElement);
					numQueuedMsgs--;
				}else{
				sendDown(qElement->getObject()->dup(), routingMsg->getOrigNode());
				}
			}
			queueIterator++;
		}
	}
	delete routingMsg;
}

//void MOFO_Net::calculateXValue(double routingMsg)


void MOFO_Net::handleLowerCTRL(MOFO_CTRL *routingMsg)
{
    ev << "received message is a CTRL message" << endl;

    statsCTRLRcvd++;

    //Check and send Queued Packets to Original CTRL msg node(destination).
    checkAndSendQueuedPktsToDest(routingMsg->getOrigNode());

    ev << "start buliding msg request" << endl;

	MOFO_RQST *my_rqst = new MOFO_RQST("RQST");
	my_rqst->setOrigNode(myAddr);


    ev << "building RQST msg blocks" << endl;
    std::vector<MOFO_MessageBlock> additional_nodes = routingMsg->getCtrlMsgTable();

    simtime_t now = simTime();

    for (unsigned int i = 0; i < additional_nodes.size(); i++)
	{
    	double offeredMsgValue = (VALOR_DA_MSG/log2(additional_nodes[i].getInitNOfCopies())*exp(LAMBDA*(SIMTIME_DBL(now) - additional_nodes[i].getMsgCreationTime())));
    	//printf ("MSG value: %f\n", offeredMsgValue);
    	cQueue::Iterator queueIterator(*queuedElements);
		QueueElement * qElement;
		bool checkMsgId = false;
		double minLocalMsgValue = VALOR_DA_MSG;
		while(!queueIterator.end()){
			qElement = dynamic_cast<QueueElement*>(queueIterator());
			if ((qElement) && ((additional_nodes[i].getMsgId()) == (qElement->getMsgId()))) {
			//if ((additional_nodes[i].getMsgId()) == (qElement->getMsgId())) {
				checkMsgId = true;
				break;
			}
			double localMsgValue = (VALOR_DA_MSG/log2(qElement->getInitNumberOfCopies())*exp(LAMBDA*(SIMTIME_DBL(now) - SIMTIME_DBL(qElement->getCreationTime()))));
			if (localMsgValue < minLocalMsgValue) minLocalMsgValue = localMsgValue;
			queueIterator++;
		}
		if ((!checkMsgId) && (numQueuedMsgs < BUFFER_SIZE_PACKETS))
			my_rqst->getRqstMsgTable().push_back(additional_nodes[i]);
		if ((!checkMsgId) && (numQueuedMsgs > BUFFER_SIZE_PACKETS) && (offeredMsgValue > minLocalMsgValue))
			my_rqst->getRqstMsgTable().push_back(additional_nodes[i]);
	}

    sendDown(my_rqst, routingMsg->getOrigNode());
	statsRQSTSent++;

	delete routingMsg;
}

void MOFO_Net::checkAndSendQueuedPktsToDest(unsigned int destinationAddress)
{
	cQueue::Iterator queueIterator(*queuedElements);
	QueueElement * qElement1;

	while(!queueIterator.end()){
		qElement1 = dynamic_cast<QueueElement*>(queueIterator());
		if ((qElement1) && (qElement1->getDestAddr() == destinationAddress)) {
			MOFO_Packet *pkt = (qElement1->getObject()->dup());
			EV << "Sending packet: ";
			statsTrafficSent++;
			pkt->setMiddleAddress(myAddr);
			sendDown(pkt, destinationAddress);
			queuedElements->remove(qElement1);
			numQueuedMsgs--;
			//return;
		}
		queueIterator++;
	}
}




void MOFO_Net::sendCTRL()
{

    ev << "Preparing CTRL broadcast " << endl;

    MOFO_CTRL *my_ctrl = new MOFO_CTRL("CTRL");

    my_ctrl->setOrigNode(myAddr);
    my_ctrl->setTimestamp(simTime());


    cQueue::Iterator queueIterator(*queuedElements);
	QueueElement * qElement;

	while(!queueIterator.end()){
		qElement = dynamic_cast<QueueElement*>(queueIterator());
		if ((qElement) && (qElement->getNumberOfCopies() >= 2)){ // && (qElement->getSrcAddr() == myAddr)) {
			MOFO_MessageBlock additional_node;
			additional_node.setMsgId(qElement->getMsgId());
			additional_node.setNOfCopies(qElement->getNumberOfCopies());
			additional_node.setInitNOfCopies(qElement->getInitNumberOfCopies());
			additional_node.setMsgCreationTime(SIMTIME_DBL(qElement->getCreationTime()));
			my_ctrl->getCtrlMsgTable().push_back(additional_node);
			//return;
		}
		queueIterator++;
	}

	std::sort(my_ctrl->getCtrlMsgTable().begin(), my_ctrl->getCtrlMsgTable().end(), compareEntries());
    sendDown(my_ctrl, LL_MANET_ROUTERS.getInt());
    statsCTRLSent++;
}


/* ------- Data Processing ------- */


void MOFO_Net::handleSelfMsg(cMessage* apMsg)
{
    ev << "handle self message" << endl;

    if (dynamic_cast<MOFO_Timeout*>(apMsg) != NULL)
    {
        // Something timed out. Let's find out what.

        if (ctrlTimeout->isExpired())
        {
        	sendCTRL();
        	ctrlTimeout->start(CTRL_MSG_PERIOD);
        }

    }
    else error("unknown message type");

}

void MOFO_Net::sendDown(cPacket* apMsg, int destAddr) {
	// WARNING: This method uses lots of hacks and/or guesswork to determine output gate and message recipient!
	// More specifically, Hosts are assumed to only have interfaces "wlan", "eth" or both. In the latter case, lowergateOut[1] is expected to be connected to the "eth" interface.

	// all messages sent to a lower layer are delayed by 0..MAXJITTER seconds (draft-ietf-manet-jitter-01)
	simtime_t jitter = dblrand() * MAXJITTER;

	// set byte size of message
	const MOFO_Packet* data = dynamic_cast<const MOFO_Packet*>(apMsg);
	const MOFO_CTRL* ctrl = dynamic_cast<const MOFO_CTRL*>(apMsg);

	/*if (ctrl) {
		apMsg->setByteLength(MOFO_CTRL_HEADER_LENGTH + (1 + ctrl->getCtrlMsgTable().size() * MOFO_MBLOCK_LENGTH));
	}
	else if (data) {

	}
	else {
		error("tried to send unsupported message type");
	}*/



	if(destAddr == (int)LL_MANET_ROUTERS.getInt()) {
		MACAddress mac_dest = MACAddress::BROADCAST_ADDRESS;
		apMsg->removeControlInfo();
		Ieee802Ctrl *control_info = new Ieee802Ctrl();
		control_info->setDest(mac_dest);
		apMsg->setControlInfo(control_info);

		if (mLowergateOut[0] != -1) sendDelayed(apMsg, jitter, mLowergateOut[0]);
		if (mLowergateOut[1] != -1) {
			cPacket* apMsg2 = (cPacket*)apMsg->dup();
			apMsg2->removeControlInfo();
			Ieee802Ctrl *control_info2 = new Ieee802Ctrl();
			control_info2->setDest(mac_dest);
			apMsg2->setControlInfo(control_info2);
			sendDelayed(apMsg2, jitter, mLowergateOut[1]);
		}

		return;
	}

	cModule* srcHost = getParentModule();
	if (!srcHost) error("source getModule(i.e. our parent) for new message not found");
	IInterfaceTable* srcIftable = check_and_cast<IInterfaceTable*>(srcHost->getSubmodule("interfaceTable"));
	if (!srcIftable) error("source module for new message has no interface table \"interfaceTable\"");
	cModule* dstHost = simulation.getModule(destAddr);
	if (!dstHost) error("destination module for new message not found");
	IInterfaceTable* dstIftable = check_and_cast<IInterfaceTable*>(dstHost->getSubmodule("interfaceTable"));
	if (!dstIftable) error("destination module for new message has no interface table \"interfaceTable\"");

	InterfaceEntry* srcWlanIf = srcIftable->getInterfaceByName("wlan");
	InterfaceEntry* srcEthIf = srcIftable->getInterfaceByName("eth");
	InterfaceEntry* dstWlanIf = dstIftable->getInterfaceByName("wlan");
	InterfaceEntry* dstEthIf = dstIftable->getInterfaceByName("eth");

	int gateIndex;
	if (srcWlanIf && !srcEthIf) {
		gateIndex = 0;
	}
	else if (!srcWlanIf && srcEthIf) {
		gateIndex = 0;
	}
	else if (srcWlanIf && srcEthIf) {
		if (dstWlanIf && !dstEthIf) {
			gateIndex = 0;
		}
		else if (!dstWlanIf && dstEthIf) {
			gateIndex = 1;
		}
		else if (dstWlanIf && dstEthIf) {
			gateIndex = 1;
		}
		else error("destination host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //0");

	}
	else error("source host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //1");

	MACAddress mac_dst;
	if (dstWlanIf && !dstEthIf) {
		mac_dst = dstWlanIf->getMacAddress();
	}
	else if (!dstWlanIf && dstEthIf) {
		mac_dst = dstEthIf->getMacAddress();
	}
	else if (dstWlanIf && dstEthIf) {
		if (srcWlanIf && !srcEthIf) {
			mac_dst = dstWlanIf->getMacAddress();
		}
		else if (!srcWlanIf && srcEthIf) {
			mac_dst = dstEthIf->getMacAddress();
		}
		else if (srcWlanIf && srcEthIf) {
			mac_dst = dstEthIf->getMacAddress();
		}
		else error("source host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //2");
	}
	else error("destination host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //3");

	apMsg->removeControlInfo();
	Ieee802Ctrl* control_info = new Ieee802Ctrl();
	control_info->setDest(mac_dst);
	apMsg->setControlInfo(control_info);

	sendDelayed(apMsg, jitter, mLowergateOut[gateIndex]);
}






void MOFO_Net::setMyAddr(unsigned int myAddr)
{
    /* Check if this node has already participated in MOFO
    if (statsRREQSent || statsRREPSent || statsRERRSent)
    {
        // TODO: Send RERRs, cold-start MOFO, lose sequence number instead?
        ev << "Ignoring IP Address change request. This node has already participated in MOFO." << endl;
        return;
    }*/

    ev << "Now assuming this node is reachable at address " << myAddr << " (was " << this->myAddr << ")" << endl;
    this->myAddr = myAddr;

    // TODO: if IInterfaceTable was autoconfigured, change IP Address there?
}


cModule* MOFO_Net::getRouterByAddress(IPAddress address) {
	return dynamic_cast<cModule*>(simulation.getModule(address.getInt() - AUTOASSIGN_ADDRESS_BASE.getInt()));
}

